function theta = learning_segmentation(varargin)

% INPUT: (all values are optional -- try at least setting modelname and boxdirname)
%    k1           ... parameter for superpixels (default = 0.08)
%    k2           ... parameter for supersegments (default = 0.16)
%    gamma        ... parameters for Pn potential (default = 1)
%    boxdirname   ... the name of the directory with LSVM detection boxes
%    modelname    ... name of the model (will be used in output files)
%    learnparams  ... parameters for learning
%    ind_classes  ... indices of classes to be used in the detection feature
%    ind_images   ... indices of images to be used in training
%    dataset_name ... name of the dataset (eg, 'mrsc' or 'voc10' (not functional yet))
% OUTPUT: learned theta

%--------------------------------------------
% SET ALL THE PATHS AND PARAMETERS
input = varargin;
[~, ~, gamma, ~, ~, modelname, learnparams, ind_classes, ind_images, params, overwrite] = parse_params(input);
ALLPATHS = learnparams.featparams.ALLPATHS;

out_dir = fullfile(ALLPATHS.LEARNING_PATH, modelname);
cache_dir = fullfile(out_dir, 'log');
if ~exist(out_dir, 'dir'), mkdir(out_dir); fileattrib(out_dir, '+w'); end;
if ~exist(cache_dir, 'dir'), mkdir(cache_dir); fileattrib(cache_dir, '+w'); end;

% check if result already exists
thetafile = fullfile(out_dir, 'theta.mat');
try %#ok<TRYNC>
   thetadata = load(thetafile);
   theta = thetadata.theta;

   fprintf('Learned theta already exists, use that and skipping training...\n');
   return;
end;

if overwrite
    delete(fullfile(cache_dir, '*.*')); 
end;

train_file = fopen(fullfile(ALLPATHS.TRAIN_FILE), 'r');
T = textscan(train_file,'%s');
imlist = T{1};
imlist = getimnames(imlist); % extract just the image name(no appendix)

if numel(ind_images)
   imlist = imlist(ind_images);
end;

% TRAINING PARAMETERS
if ~numel(learnparams) % take default values
    p = 2;
    C = 2;
    epsilon = 1;
    rGap = 1e-5;  %the gap between Primal and Dual
    it = 10000;    %the Number of Iterations 
    learnparams.p = p;
    learnparams.C = C;
    learnparams.epsilon = epsilon;
    learnparams.rGap = rGap;
    learnparams.it = it;
else
    p = learnparams.p;
    C = learnparams.C;
    epsilon = learnparams.epsilon;
    rGap = learnparams.rGap;
    it = learnparams.it;
end
if ~isfield(learnparams, 'featparams')
    featparams = defaultfeatparams([], ind_classes);
    learnparams.featparams = featparams;
end;

featparams = defaultfeatparams(learnparams.featparams, ind_classes);  % if values missing, fill default
learnparams.featparams = featparams;

%-------------------------
% which classes (ind_classes) are to be used in detection
if numel(ind_classes)
    ind_classes_det = load(fullfile(ALLPATHS.BBOX_PATH, 'classinfo.txt'));
    ind = intersect(ind_classes_det, ind_classes);
    if length(ind) < length(ind_classes)   % check that all wanted classes are really there
        fprintf(['Detection only available for classes: ' repmat('%d ', [1, length(ind)])], ind);
        fprintf(' using ...\n');
        pause(5);
        ind_classes = ind;
    end;
end;
%-------------------------

NumStates = ALLPATHS.NUM_STATES;
IGNORE_CLASSES = ALLPATHS.IGNORE_CLASSES;
BinaryStates = 2;   %the number of states for 'Zi'
params.learnparams = learnparams;
params.NumStates = NumStates;
params.BinaryStates = BinaryStates;
params.imlist = imlist;
params.ignore_classes = ALLPATHS.IGNORE_CLASSES;
train_params_file = fullfile(out_dir, 'params.mat'); % write the parameters for learning
save(train_params_file, 'params')
%--------------------------------------------


%--------------------------------------------
% Compute (or load -- for debugging purposes) the features
tic;
featfile = fullfile(cache_dir, 'features-train.mat');
try
    load(featfile); 
catch
    [features, empmean, CRFDATA] = computeFeatures(imlist, NumStates, IGNORE_CLASSES, ind_classes, gamma, cache_dir, ALLPATHS, 1, featparams);
    save(featfile, 'features', 'empmean', 'CRFDATA');
end;
featcomputetime = toc;
%--------------------------------------------


%--------------------------------------------
% Compute (or load) the Loss Function
tic;
lossfile = fullfile(cache_dir, 'loss-train.mat');
loss = computeLoss(imlist, CRFDATA, IGNORE_CLASSES, ind_classes, ALLPATHS);
save(lossfile, 'loss');
losscomputetime = toc;
%--------------------------------------------


%--------------------------------------------
% TRAINING
NumFeatures = length(features);
th_init = ones(NumFeatures,1);

disp('------Training------');
tic;
theta = libCRFmatlab(features, loss, th_init, epsilon, 1, empmean, p, C, rGap, it);
disp('------theta------');
disp(theta);

% write results
traintime = toc;
thetafile = fullfile(out_dir, 'theta.mat');
save(thetafile, 'theta')
fprintf('finished!\n');
logfile = fullfile(out_dir, 'log.txt');
fid = fopen(logfile, 'w+');
fprintf(fid, 'Num train images: %d\n', length(imlist));
fprintf(fid, 'Num classes for detection: %d\n', length(ind_classes));
fprintf(fid, 'Num all classes: %d\n', NumStates);
fprintf(fid, 'Feature compute time: %0.6f\n', featcomputetime);
fprintf(fid, 'Loss compute time: %0.6f\n', losscomputetime);
fprintf(fid, 'Train time: %0.6f', traintime);
fclose(fid);
fprintf('Training time: %0.6f sec\n', traintime)
%---------------------------------------


%---------------------------------------
% get parameters
function [k1, k2, gamma, NumFeatures, boxdirname, modelname, learnparams, ind_classes, ind_images, params, overwrite] = parse_params(input)

% put default values
k1 = 0.08;
k2 = 0.16;
gamma = 1;
NumFeatures = 6;
boxdirname = 'LSVM-seg-gt';   % default is groundtruth
modelname = 'model';
learnparams = [];
datasetname = 'msrc';
ind_classes = 7;
ind_images = [];
overwrite = 0;

for i = 1 : 2 : length(input)  % every second entry is the value of the variable
   label = input{i};
   switch label
       case 'k1'
           k1 = input{i+1};
       case 'k2'
           k2 = input{i+1};
       case 'gamma'
           gamma = input{i+1};
       case 'NumFeatures'
           NumFeatures = input{i+1};
       case 'boxdirname'
           boxdirname = input{i+1};
       case 'modelname'
           modelname = input{i+1};
       case 'learnparams'
           learnparams = input{i+1};
       case 'datasetname'
           datasetname = input{i+1};
       case 'ind_classes'
           ind_classes = input{i+1};
       case 'ind_images'
           ind_images = input{i+1};
   end;
end
params.k1 = k1;
params.k2 = k2;
params.gamma = gamma;
params.NumFeatures = NumFeatures;
params.boxdirname = boxdirname;
params.modelname = modelname;
params.learnparams = learnparams;
params.datasetname = datasetname;
params.ind_classes = ind_classes;
%---------------------------------------